"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Pos = exports.dedent = exports.unexp = exports.isObj = exports.ZamlError = exports.basicTypeFromName = exports.BLOCKABLE_TYPES = exports.LISTABLE_TYPES = exports.validTypes = exports.trailingWhitespace = exports.reservedOps = exports.whitespace = exports.brackets = void 0;
//
// Constants
//
exports.brackets = /[\(\)\{\}\[\]]/;
exports.whitespace = /[ \t\n\r]/;
exports.reservedOps = /[\^&#@~`]/;
exports.trailingWhitespace = /[ \t\n\r]*$/;
exports.validTypes = ['num', 'str', 'kv', 'list', 'bool', 'enum'];
exports.LISTABLE_TYPES = ['str', 'num', 'enum'];
exports.BLOCKABLE_TYPES = ['str', 'num', 'bool', 'tuple'];
function basicTypeFromName(name) {
    if (name === 'num' || name === 'str' || name === 'bool') {
        return { type: name };
    }
    return null;
}
exports.basicTypeFromName = basicTypeFromName;
class ZamlError extends Error {
    constructor(type, pos, message) {
        super(`[ZamlError] ${message}`);
        this.type = type;
        this.pos = pos;
    }
    static syntax(pos, char, more = '') {
        return new ZamlError('syntax-error', pos, `Unexpected '${char}'${more}`);
    }
    addSource(source) {
        if (!this.pos)
            return;
        var lines = source.split('\n');
        var { line, col } = this.pos;
        var lineNo = String(line);
        var prevLineNo = String(line - 1);
        if (lineNo.length > prevLineNo.length) {
            prevLineNo = spacer(lineNo.length - prevLineNo.length) + prevLineNo;
        }
        else if (prevLineNo.length > lineNo.length) {
            lineNo = spacer(prevLineNo.length - lineNo.length) + lineNo;
        }
        this.message =
            this.message + '\n' +
                `  on line ${line} col ${col}\n\n` +
                (line > 1 ? `  ${prevLineNo}│ ${lines[line - 2]}\n` : '') +
                `  ${lineNo}│ ${lines[line - 1]}\n  ${spacer(col - 1 + lineNo.length + 2)}^`;
    }
}
exports.ZamlError = ZamlError;
//
// Helpers
//
function isObj(x) {
    return Object.prototype.toString.call(x) === '[object Object]';
}
exports.isObj = isObj;
function unexp(char, more = '') {
    return `Unexpected ${JSON.stringify(char)}${more}`;
}
exports.unexp = unexp;
function range(start, end) {
    var result = [];
    for (var i = start; i < end; i++)
        result.push(i);
    return result;
}
function spacer(length) {
    return range(0, length).map(_ => ' ').join('');
}
// Taken and modified from https://github.com/dmnd/dedent
function dedent(string) {
    const lines = string.split("\n");
    let mindent = 0;
    lines.forEach(l => {
        let m = l.match(/^(\s+)\S+/);
        if (m) {
            let indent = m[1].length;
            if (!mindent) {
                // this is the first indented line
                mindent = indent;
            }
            else {
                mindent = Math.min(mindent, indent);
            }
        }
    });
    let result = '';
    if (mindent !== null) {
        const m = mindent; // appease TS
        result = lines.map(l => l[0] === " " ? l.slice(m) : l).join("\n");
    }
    return result
        // dedent eats leading and trailing whitespace too
        .trim()
        // handle escaped newlines at the end to ensure they don't get stripped too
        .replace(/\\n/g, "\n");
}
exports.dedent = dedent;
class Pos {
    constructor(line = 1, col = 1, i = 0) {
        this.line = line;
        this.col = col;
        this.i = i;
    }
    push(source) {
        this.skipNewline(source) || this.newcol();
        return this;
    }
    newcol() {
        this.i += 1;
        this.col += 1;
        return this;
    }
    newline(windowsLine = false) {
        this.i += (windowsLine ? 2 : 1);
        this.col = 1;
        this.line += 1;
    }
    copy() {
        return new Pos(this.line, this.col, this.i);
    }
    goto(target) {
        this.i = target.i;
        this.col = target.col;
        this.line = target.line;
    }
    skipSpace(source) {
        if (source[this.i] === ' ' || source[this.i] === '\t') {
            this.newcol();
            return true;
        }
        return false;
    }
    skipNewline(source) {
        if (source[this.i] === '\n') {
            this.newline(source[this.i + 1] === '\r');
            return true;
        }
        return false;
    }
    skipWhitespace(source) {
        if (this.i >= source.length)
            return false;
        return this.skipSpace(source) || this.skipNewline(source);
    }
}
exports.Pos = Pos;
